import sys

from typing import Dict, Optional, IO

from threading import Thread, Event

from logging import Logger


class LogStream(object):
    """
    Reads from a stream in a background thread and logs the result line by line
    """

    def __init__(
        self, logger: Logger, log_level: int, stream: IO[bytes], extra: Dict[str, str]
    ) -> None:
        """
        :param Logger logger: the logger to which the stream content is written
        :param int log_level: the log level (see Python logging)
        :param IOBase stream: the stream that is logged
        :param Dict[str, str] extra: extra information that is logged each line (see Python logging)
        """
        self.logger = logger
        self.log_level = log_level
        self.stream = stream
        self.extra = extra
        self.thread: Optional[Thread] = None
        self._closed = Event()

    def _run(self):
        try:
            for line in self.stream:
                strline = line.decode(sys.getdefaultencoding()).rstrip("\r\n\t ")
                self.logger.log(self.log_level, strline, extra=self.extra)
        except ValueError:
            pass  # stream was closed
        except OSError:
            self.logger.exception(
                "I/O Error while logging: %s", self.name, extra=self.extra  # type: ignore
            )
        except:
            self.logger.exception(
                "Something went wrong while logging", extra=self.extra
            )
            raise
        finally:
            self.stream.close()
            self._closed.set()

    def start(self):
        """starts reading and logging"""
        program = self.extra.get("program", "")
        name = f"{program}:{self.log_level}"
        thread = Thread(target=lambda: self._run(), name=name)
        thread.daemon = True
        self.thread = thread
        thread.start()
        return self

    def wait_close(self, wait_time: float = 1.0):
        """
        waits untill the stream closes or a timeout occurs

        :param wait_time: timeout in seconds, defaults to 1.0
        :type wait_time: float, optional
        """
        try:
            if self.thread:
                self._closed.wait(wait_time)
        except IOError:
            pass



import yaml
import marshmallow_dataclass

from typing import Dict, List, Any, Optional, Union
from logging import Logger, getLogger, INFO, getLevelName
from pluggy import HookimplMarker  # type: ignore

from dataclasses import dataclass
from marshmallow.exceptions import MarshmallowError, ValidationError
from io import TextIOBase

import re
from datetime import datetime

import os
import stat
from threading import Thread, Event

ENCAB = "encab"
LOG_COLLECTOR = "log_collector"

mylogger = getLogger(LOG_COLLECTOR)


class ConfigError(ValueError):
    pass


@dataclass
class SourceConfig(object):

    path: Optional[str]
    """The path of the log file"""

    path_pattern: Optional[str]
    """
    The path pattern of the log file
    
    ``%(<name>)e`` inserts the value of the environment variable with name <name>
    
    ``%(<dateformat>)d`` inserts the current time with <dateformat> as python date format, 
    
    see https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes
    
    ``%`` must be masked with ``%%``
    
    examples:
    
    - ``%(HOME)e/path`` -> ``/home/user/path``
    - ``error-%(%y%m%d)d.log`` -> ``error-20230103.log``
    - ``a%%b.log`` -> ``a%b.log``
    """

    offset: Optional[int]
    """The initial read offset, defaults to 0
    
       -1     : Start at the beginning of the file
        0     : Start at the end of the file
        n > 0 : Start n characters before the end of the file
    """

    level: Union[str, int, None]
    """The log level"""

    poll_interval: Optional[float]
    """The poll interval in seconds"""

    def _set_log_level(self):
        """
        checks the log level and turns it into an int if necessary

        :raises ConfigError: if an unspecified log level is given
        """
        levels = ["CRITICAL", "FATAL", "ERROR", "WARN", "WARNING", "INFO", "DEBUG"]

        level = self.level

        if level and isinstance(level, str) and level not in levels:
            levels_printed = ", ".join(levels)
            raise ConfigError(
                f"Unsupported log level {level}. Supported levels are: {levels_printed}"
            )

        level = level or INFO
        self.level = level if isinstance(level, int) else getLevelName(level)

    def __post_init__(self):
        if not self.path and not self.path_pattern:
            raise ConfigError(
                "Either path ot path_pattern must be set in log_collector source"
            )

        if self.path and self.path_pattern:
            raise ConfigError(
                "Either path ot path_pattern must be set but not both in log_collector source"
            )

        self._set_log_level()


@dataclass
class LogCollectorSettings(object):
    """
    the log collector settings

    eample:

    .. code-block:: yaml

        encab:
            debug: true
            halt_on_exit: False
        extensions:
            log_collector:
                enabled: true
                settings:
                    sources:
                        error_log:
                            file: ./error.log

    This class contains the extensions/log_collector/settings content.
    """

    sources: Dict[str, SourceConfig]

    def __post_init__(self):
        pass

    @staticmethod
    def load(settings: Dict[str, Any]) -> "LogCollectorSettings":
        try:
            ConfigSchema = marshmallow_dataclass.class_schema(LogCollectorSettings)
            return ConfigSchema().load(settings)  # type: ignore
        except ValidationError as e:
            msg = e.args[0]
            if isinstance(msg, dict):
                msg = yaml.dump(msg, default_flow_style=False)

            raise ConfigError(f"\n\n{LOG_COLLECTOR}:\n{msg}")
        except MarshmallowError as e:
            raise ConfigError(e.args)


class PathPattern(object):
    FORMAT = re.compile(r"((%%|[^%])*)|(%\([^\)]*\)[ed])")

    def __init__(self, pattern: str) -> None:
        self.tokens = self.FORMAT.findall(pattern)

    def replace(self, pattern: str, time: datetime, environment: Dict[str, str]) -> str:
        suffix = pattern[-1]
        pattern = pattern[2:-2]

        if suffix == "e":
            name = pattern
            if name not in environment:
                return ""

            return environment[name]
        else:
            date_format = pattern
            try:
                return time.strftime(date_format)
            except ValueError as e:
                raise ConfigError(
                    f"Malformed date format '{date_format}' in log_collector source: {str(e)}"
                )

    def format(self, time: datetime, environment: Dict[str, str]) -> str:
        result = list()
        for token in self.tokens:
            if token[0]:
                result.append(token[0].replace("%%", "%"))
            elif token[2]:
                result.append(self.replace(token[2], time, environment))

        return "".join(result)


class LogPath(object):
    def __init__(
        self, path_or_pattern: str, environment: Dict[str, str], fixed: bool
    ) -> None:
        self._fixed_path = path_or_pattern if fixed else None
        self._path_pattern = None if fixed else PathPattern(path_or_pattern)
        self._environment = environment

    @staticmethod
    def fixed(path: str) -> "LogPath":
        return LogPath(path, dict(), True)

    @staticmethod
    def variable(path_pattern: str, environment: Dict[str, str]) -> "LogPath":
        return LogPath(path_pattern, environment, False)

    def current(self, time: datetime):
        if self._fixed_path:
            return self._fixed_path
        else:
            time = time or datetime.now()
            assert self._path_pattern
            return self._path_pattern.format(time, self._environment)

    def is_fixed(self):
        return self._fixed_path is not None


class Stopped(Exception):
    pass


class LogCollector(object):
    def __init__(
        self,
        name: str,
        path: LogPath,
        logger: Logger,
        level: int = INFO,
        offset: int = 0,
        poll_interval: float = 0.5,
    ) -> None:
        self.name = name
        self.path = path
        self.logger = logger
        self.level = level
        self.extra = {"program": name}
        self.offset = offset
        self.poll_interval = poll_interval

        self._thread = None
        self._started = Event()
        self._stop = Event()
        self._stopped = Event()
        self._current_time: Optional[datetime] = None
        self._fp: Optional[TextIOBase] = None
        self._file_existed_at_start = False

    def __now(self) -> datetime:
        return self._current_time or datetime.now()

    def current_path(self) -> str:
        return self.path.current(self.__now())

    def __st_mode(self) -> int:
        return os.stat(self.current_path()).st_mode

    def file_exists(self):
        try:
            st_mode = self.__st_mode()
            return stat.S_ISREG(st_mode) or stat.S_ISFIFO(st_mode)
        except FileNotFoundError:
            return False

    def is_regular_file(self):
        try:
            st_mode = self.__st_mode()
            return stat.S_ISREG(st_mode)
        except FileNotFoundError:
            return False

    def check_stopped(self):
        if self._stop.is_set():
            raise Stopped()

    def get_fp(self):
        if self._fp:
            return self._fp
        else:
            raise Stopped()

    def clear_fp(self):
        self._fp = None

    def log_lines(self):
        for line in self.get_fp():
            line = line.rstrip("\r\n\t ")
            self.logger.log(self.level, line, extra=self.extra)

    def open(self, path: str) -> TextIOBase:
        fp = open(path, "r")
        self._fp = fp
        return fp

    def fast_forward(self):
        if self.offset == -1 or not self._file_existed_at_start:
            return

        fp = self.get_fp()
        fp.seek(0, 2)
        pos = fp.tell()
        pos = max(pos - self.offset, 0)

        if pos > 0:
            fp.seek(pos)

    def poll_data(self, path: str):
        fp = self.get_fp()
        pos = fp.tell()
        self.logger.debug(
            f"Waiting for new data in file {path}...",
            extra=self.extra,
        )
        self._stop.wait(self.poll_interval)
        self.check_stopped()
        fp.seek(pos)

    def collect_fifo(self):
        path = self.current_path()
        with self.open(path):
            self.log_lines()

    def collect_file(self):
        path = self.current_path()
        with self.open(path):
            self.fast_forward()
            while True:
                self.log_lines()
                self.poll_data(path)

    def same_log_file(self, path: str) -> bool:
        current_path = self.current_path()
        same_file = path == current_path

        if not same_file:
            mylogger.debug("Log path changed from %s to %s", path, current_path)

        return same_file

    def collect_rolling_files(self):
        is_first_file = True

        while True:
            path = self.current_path()
            with self.open(path):
                if is_first_file:
                    self.fast_forward()
                    is_first_file = False

                while True:
                    self.log_lines()
                    self.check_stopped()

                    if not self.same_log_file(path):
                        break

                    self.poll_data(path)

    def poll_file(self):
        while not self.file_exists():
            self.check_stopped()
            self.logger.debug(
                "Waiting for file %s ...", self.current_path(), extra=self.extra
            )
            self._stop.wait(self.poll_interval)

    def _run(self):
        try:
            self._file_existed_at_start = self.file_exists()
            self._started.set()

            while True:
                try:
                    self.clear_fp()
                    self.poll_file()
                    self.check_stopped()
                    if not self.file_exists():
                        raise Stopped()

                    if not self.is_regular_file():
                        self.collect_fifo()
                    else:
                        if self.path.is_fixed():
                            self.collect_file()
                        else:
                            self.collect_rolling_files()

                except FileNotFoundError:
                    pass
                except BrokenPipeError:
                    pass

        except Stopped:
            pass
        except Exception as e:
            self.logger.exception(
                "Reading source %s failed: %s", self.name, str(e), extra=self.extra
            )
        finally:
            self.clear_fp()
            self._stopped.set()

    def start(self, wait_time: float = 0.1, current_time: Optional[datetime] = None):
        self._current_time = current_time
        thread = Thread(target=lambda: self._run(), name=self.name)
        thread.daemon = True
        self.thread = thread
        thread.start()
        self._started.wait(wait_time)
        return self

    def set_current_time(self, current_time: datetime):
        self._current_time = current_time

    def stop(self, wait_time: float = 1.0):
        self._stop.set()
        if self._fp:
            try:
                self._fp.close()
            except IOError:
                pass
            self._fp = None
        self._stopped.wait(wait_time)


extension_impl = HookimplMarker(ENCAB)


class LogCollectorExtension(object):
    def __init__(self) -> None:
        self.settings: Optional[LogCollectorSettings] = None
        self.enabled = True
        self.collectors: List[LogCollector] = list()

    def add_collectors(self, environment: Dict[str, str]):
        if not self.settings:
            return

        for name, source in self.settings.sources.items():
            path = None
            if source.path:
                path = LogPath.fixed(source.path)
            if source.path_pattern:
                path = LogPath.variable(source.path_pattern, environment)

            assert path
            level = source.level
            assert isinstance(level, int)
            collector = LogCollector(
                name=name,
                path=path,
                logger=mylogger,
                level=level,
                offset=source.offset or -1,
                poll_interval=source.poll_interval or 1,
            )
            self.collectors.append(collector)

    def start_collectors(self):
        for collector in self.collectors:
            collector.start()

    def stop_collectors(self):
        for collector in self.collectors:
            collector.stop()

    @extension_impl
    def validate_extension(self, name: str, enabled: bool, settings: Dict[str, Any]):
        if name == LOG_COLLECTOR:
            self.settings = LogCollectorSettings.load(settings)
            mylogger.info("settings are valid.", extra={"program": ENCAB})

    @extension_impl
    def configure_extension(self, name: str, enabled: bool, settings: Dict[str, Any]):
        if name != LOG_COLLECTOR:
            return

        if not enabled:
            self.enabled = False
            return

        self.settings = LogCollectorSettings.load(settings)

    @extension_impl
    def extend_environment(self, program_name: str, environment: Dict[str, str]):
        if not self.enabled:
            return

        if not self.collectors:
            self.add_collectors(environment)
            self.start_collectors()

    @extension_impl
    def programs_ended(self):
        if self.enabled:
            self.stop_collectors()



import re
import yaml
import marshmallow_dataclass

from re import match, compile

from abc import ABC, abstractmethod
from typing import Dict, Set, List, Any, Optional, Union
from logging import getLogger
from pluggy import HookimplMarker  # type: ignore

from yaml.error import YAMLError
from dataclasses import dataclass
from marshmallow.exceptions import MarshmallowError, ValidationError

ENCAB = "encab"
VALIDATION = "validation"

mylogger = getLogger(VALIDATION)


class ConfigError(ValueError):
    pass


@dataclass
class Validation(object):

    required: Optional[bool]
    """True: this variable is required"""

    format: Optional[str]
    """Variable format. One of ``string``, ``float``, ``int``. Default: ``string``  """

    default: Union[str, int, float, None]
    """Default value of this variable. It will be set if this variable is missing or empty"""

    min_length: Optional[int]
    """Minimum length"""

    max_length: Optional[int]
    """Maximum length"""

    min_value: Union[int, float, None]
    """Minimum value"""

    max_value: Union[int, float, None]
    """Maximum value"""

    regex: Optional[str]
    """
    If set, the value must match the Regular expression given.
    See https://docs.python.org/3/howto/regex.html. 
    """

    program: Optional[str]
    """
    Validation is limited to the given program. Default: no limitation.
    
    Use ``programs`` if validation should be limited to multiple programs.
    """

    programs: Optional[List[str]]
    """
    Validation is limited to the given programs. Default: no limitation.
    """

    def _set_range(self):
        min_value = self.min_value
        max_value = self.max_value

        if min_value and max_value and max_value < min_value:
            raise ConfigError(
                f"Expected max_value >= min_value was {max_value} < {min_value}"
            )

    def _set_programs(self):
        if self.program and self.programs:
            raise ConfigError("Expected either 'program' or 'programs' but got both.")

        self.programs = [self.program] if self.program else (self.programs or list())

    def _set_length(self):
        min_length = self.min_length
        max_length = self.max_length

        if min_length and min_length < 0:
            raise ConfigError(f"Expected min_length >=0 but was {min_length}")

        if max_length and max_length < 0:
            raise ConfigError(f"Expected max_length >=0 but was {max_length}")

        if min_length and max_length and max_length < min_length:
            raise ConfigError(
                f"Expected max_length >= min_length was {max_length} < {min_length}"
            )

    def _set_format(self):
        self.format = self.format or "string"

        SUPPORTED_FORMATS = ("string", "float", "int")

        if self.format not in SUPPORTED_FORMATS:
            formats = ", ".join(SUPPORTED_FORMATS)
            raise ConfigError(
                f"Unsupported float format. Supported formats are: {formats}"
            )

    def _set_regex(self):
        regex = self.regex
        if isinstance(regex, str):
            try:
                compile(regex)
            except ValueError as e:
                raise ConfigError(
                    f"Expected 'regex' to be a valid regex but was {regex}: {str(e)}"
                )

    def __post_init__(self):
        self.required = True if self.required is None else self.required

        self._set_regex()
        self._set_format()
        self._set_length()
        self._set_programs()
        self._set_range()


@dataclass
class ValidationSettings(object):
    """
    the validation settings

    eample:

    .. code-block:: yaml

        encab:
            debug: true
            halt_on_exit: False
        extensions:
            validation:
                enabled: true
                settings:
                    include: "validation.yml"
                    variables:
                        X:
                            required: true
                            default: "1"
                            min_length: 1
                            max_length: 5
                            regex: "0|1"
                        Y:
                            min_value: 0
                            max_value: 10


    This class contains the extensions/variables/settings content.
    """

    variables: Optional[Dict[str, Validation]]
    """ the environment variable specifications """

    include: Optional[str]
    """ include additional environment variable specifications from file 
    
        example settings:
        
        .. code-block:: yaml
        
            settings:
                include: validation.yml
                    
        example file ``validation.yml``:
        
        .. code-block:: yaml
        
            X:
                required: true
                default: "1"
                min_length: 1
                max_length: 5
                regex: "0|1"
            Y:
                min_value: 0
                max_value: 10
        
    """

    def __post_init__(self):
        self.variables = self.variables or {}

    @staticmethod
    def load(settings: Dict[str, Any]) -> "ValidationSettings":
        try:
            ConfigSchema = marshmallow_dataclass.class_schema(ValidationSettings)
            return ConfigSchema().load(settings)  # type: ignore
        except ValidationError as e:
            msg = e.args[0]
            if isinstance(msg, dict):
                msg = yaml.dump(msg, default_flow_style=False)

            raise ConfigError(f"\n\n{VALIDATION}:\n{msg}")
        except MarshmallowError as e:
            raise ConfigError(e.args)

    def include_validations(self) -> Dict[str, Validation]:
        """
        includes validations from an external yaml file

        :raises ConfigError: if the validation file could not be loaded
        :return: a dictionary of validations
        :rtype: Dict[str, Validation]
        """
        validations: Dict[str, Validation] = dict()
        if not self.include:
            return validations

        prefix = f"{VALIDATION}: Failed to include {self.include}"
        try:
            with open(self.include, "r") as stream:
                map = yaml.safe_load(stream)

            if not isinstance(map, dict):
                raise ConfigError(
                    f"{prefix}: Expected map as root element but got {str(map)}"
                )

            for var, validation_map in map.items():
                if not isinstance(var, str):
                    raise ConfigError(f"{prefix}: Invalid variable name {var}")

                ConfigSchema = marshmallow_dataclass.class_schema(Validation)
                validation = ConfigSchema().load(validation_map)
                assert isinstance(validation, Validation)
                validations[var] = validation

            return validations
        except YAMLError as e:
            raise ConfigError(f"{prefix}: YAML error(s) {str(e)}")
        except ValidationError as e:
            msg = e.args[0]
            if isinstance(msg, dict):
                msg = yaml.dump(msg, default_flow_style=False)

            raise ConfigError(f"{prefix}.\n\n{VALIDATION}:\n{msg}")
        except MarshmallowError as e:
            raise ConfigError(f"{prefix}:" + str(e.args[0]), e.args[1:])


extension_impl = HookimplMarker(ENCAB)


class VariableValidator(ABC):
    def __init__(self, name: str, validation: Validation) -> None:
        self.name = name
        self.validation = validation

    @abstractmethod
    def validate(self, value: str):
        pass

    def report_error(self, message: str) -> ConfigError:
        return ConfigError(
            f"{VALIDATION}: Validation for variable {self.name} failed: {message}."
        )


class FormatValidator(VariableValidator):
    def validate(self, value: str):
        format = self.validation.format

        if format == "string":
            pass
        elif format == "int":
            try:
                int(value)
            except ValueError:
                raise self.report_error("Expected integer format but was '{value}'")
        elif format == "float":
            try:
                float(value)
            except ValueError:
                raise self.report_error("Expected float format but was '{value}'")
        else:
            assert False, f"Unsupported format {format}."


class RangeValidator(VariableValidator):
    def validate(self, value: str):
        min_value = self.validation.min_value
        max_value = self.validation.max_value

        if not (min_value or max_value):
            return

        try:
            n = float(value)
        except ValueError:
            raise self.report_error(
                "Expected to be float (as min_value and/or max_value was given)"
                f" but was '{value}'"
            )

        if min_value and n < min_value:
            raise self.report_error("Expected {self.name} >= {min_value} but was {n}")

        if max_value and n > max_value:
            raise self.report_error(
                f"{VALIDATION}: Expected {self.name} <= {max_value} but was {n}"
            )


class LengthValidator(VariableValidator):
    def validate(self, value: str):
        min_length = self.validation.min_length
        max_length = self.validation.max_length

        if min_length and len(value) < min_length:
            raise self.report_error(
                f"Expected length >= {min_length} but was {len(value)}"
            )

        if max_length and len(value) > max_length:
            raise self.report_error(
                f"Expected length <= {max_length} but was {len(value)}"
            )


class RegexValidator(VariableValidator):
    def validate(self, value: str):
        regex = self.validation.regex

        if regex and not match(regex, value):
            raise self.report_error(f"Expected to match '{str(regex)}'")


class CombinedValidator(VariableValidator):
    def __init__(self, name: str, validation: Validation) -> None:
        super().__init__(name, validation)
        validator_classes = (
            FormatValidator,
            RangeValidator,
            LengthValidator,
            RegexValidator,
        )
        self.validators = [c(name, validation) for c in validator_classes]  # type: ignore

    def validate(self, value: str):
        for validator in self.validators:
            validator.validate(value)


class Validator(object):
    def __init__(self) -> None:
        self.settings = ValidationSettings(dict(), None)
        self.validations: Dict[str, Validation] = dict()

    def validate_names(self):
        pattern = re.compile(r"^[a-zA-Z_]+[a-zA-Z0-9_]*")
        for name in self.validations.keys():
            if not pattern.match(name):
                raise ConfigError(
                    f"{VALIDATION}: Expected valid environment variable name (see POSIX 3.231 Name)"
                    f" but was '{name}'."
                )

    def update_settings(self, settings: ValidationSettings):
        validations = settings.include_validations()
        self.validations.update(validations)
        assert settings.variables
        self.validations.update(settings.variables)
        self.validate_names()

    def validate(self, program: str, name: str, value: str):
        validation: Optional[Validation] = self.validations.get(name)

        if not validation:
            return

        if validation.programs and program not in validation.programs:
            return

        mylogger.debug(
            "Validating variable %s for program %s",
            name,
            program,
            extra={"program": ENCAB},
        )

        CombinedValidator(name, validation).validate(value)

    def validate_all(self, program: str, vars: Dict[str, str]):
        vars_set: Set[str] = set()

        for name, value in vars.items():
            if value:
                vars_set.add(name)
                self.validate(program, name, value)

        for name, validation in self.validations.items():

            if validation.programs and program not in validation.programs:
                continue

            if name in vars_set:
                continue

            if validation.default:
                mylogger.debug(
                    "Program %s: setting default %s = %s",
                    program,
                    name,
                    validation.default,
                    extra={"program": ENCAB},
                )
                vars[name] = str(validation.default)
                continue

            if validation.required:
                raise ConfigError(f"{VALIDATION}: Missing required parameter {name}")


class ValidationExtension(object):
    """
    Validates environment variables, sets default values and stops program execution if one or more variables
    do not conform to the specification.
    """

    def __init__(self) -> None:
        self.validator = Validator()
        self.enabled = True
        self.programs_updated: Set[str] = set()

    @extension_impl
    def validate_extension(self, name: str, enabled: bool, settings: Dict[str, Any]):
        if name == VALIDATION:
            self.validator.update_settings(ValidationSettings.load(settings))
            mylogger.info("settings are valid.", extra={"program": ENCAB})

    @extension_impl
    def configure_extension(self, name: str, enabled: bool, settings: Dict[str, Any]):
        if name != VALIDATION:
            return

        if not enabled:
            self.enabled = False
            return

        self.validator.update_settings(ValidationSettings.load(settings))

    @extension_impl
    def extend_environment(self, program_name: str, environment: Dict[str, str]):
        if not self.enabled:
            return

        if program_name == ENCAB:
            return

        if program_name in self.programs_updated:
            self.validator.validate_all(program_name, environment)
        else:
            self.programs_updated.add(program_name)



import os
import sys
import re
import yaml
import marshmallow_dataclass

from io import StringIO
from typing import Dict, List, Any, Optional, Union
from logging import getLogger
from pluggy import HookimplMarker  # type: ignore

from dataclasses import dataclass
from marshmallow.exceptions import MarshmallowError, ValidationError

from dotenv import dotenv_values
from subprocess import Popen

from encab.common.process import Process

ENCAB = "encab"
STARTUP_SCRIPT = "startup_script"

mylogger = getLogger(STARTUP_SCRIPT)


class ConfigError(ValueError):
    pass


@dataclass
class StartupScriptSettings(object):
    """
    the startup script settings

    eample:

    .. code-block:: yaml

        encab:
            debug: true
            halt_on_exit: False
        extensions:
            startup_script:
                enabled: true
                settings:
                    loadenv:
                        .env
                    buildenv:
                        - echo "X=1"
                    sh:
                        - echo $X

    This class contains the extensions/startup_script/settings content.
    """

    loadenv: Optional[str]
    """
    the path of the file in dotenv format from which environment variables are loaded.
    The format is similar to Bash files.
    
    Example:
    
    .. code-block:: bash
    
        FOO="first line"
        BAR=2
    
    see: https://pypi.org/project/python-dotenv/
    
    """

    buildenv: Union[List[str], str, None]
    """
    executes a shell command that generates environment variables in dotenv format.
    
    Example:
    
    .. code-block:: yaml
    
        buildenv:
        - echo "X=1"
        - echo "Y=2"
    """

    sh: Union[List[str], str, None]
    """
    executes a shell command before the programs specified in the programs section of the encab file are run
    """

    def __post_init__(self):
        if isinstance(self.buildenv, str):
            self.buildenv = [self.buildenv]

        if isinstance(self.sh, str):
            self.sh = [self.sh]

    @staticmethod
    def load(settings: Dict[str, Any]) -> "StartupScriptSettings":
        try:
            ConfigSchema = marshmallow_dataclass.class_schema(StartupScriptSettings)
            return ConfigSchema().load(settings)  # type: ignore
        except ValidationError as e:
            msg = e.args[0]  # type: ignore
            if isinstance(msg, dict):
                msg = yaml.dump(msg, default_flow_style=False)

            raise ConfigError(f"\n\n{STARTUP_SCRIPT}:\n{msg}")
        except MarshmallowError as e:
            raise ConfigError(e.args)


class StartupScript:
    """
    Run scripts before the actual programs are started.
    In addition, environment variables can be loaded from a file or generated
    using a script.
    """

    def __init__(self) -> None:
        self.settings: Optional[StartupScriptSettings] = None
        self.executed = False

    def update_settings(self, settings: StartupScriptSettings):
        self.settings = settings

    def loadenv(self, environment: Dict[str, str]):
        """
        loads the environment from a file if the plugin settings demand it

        :param environment: the environment
        :type environment: Dict[str, str]
        :raises ConfigError: if the file could not be loaded
        """
        if not self.settings:
            return

        path = self.settings.loadenv

        if not path:
            return

        if not os.path.isfile(path):
            raise ConfigError(
                f"{STARTUP_SCRIPT}: File {path} defined by loadenv does not exist"
            )

        mylogger.info("Loading env file: %s", path, extra={"program": ENCAB})

        self.update_env(environment, path=path)

    def update_env(
        self,
        environment: Dict[str, str],
        path: Optional[str] = None,
        stream: Optional[StringIO] = None,
    ):
        """
        validates and updates the environment form a file or stream in dotenv_ format

        .. _dotenv: https://pypi.org/project/python-dotenv/

        :param environment: the environment
        :type environment: Dict[str, str]
        :param path: the path from which the environment should be loaded, defaults to None
        :type path: Optional[str], optional
        :param stream: _description_, defaults to None
        :type stream: Optional[StringIO], optional
        :raises IOError: if the environment could not be loaded
        """
        try:
            if path:
                values = dotenv_values(dotenv_path=path)
            else:
                assert stream
                values = dotenv_values(stream=stream)

            if isinstance(values, dict):
                env = self.clean_up_env(values)
                mylogger.debug(
                    "Adding environment: %s", str(env), extra={"program": ENCAB}
                )
                environment.update(env)

        except IOError:
            raise IOError(
                f"{STARTUP_SCRIPT}: Failed to load environment specified in loadenv from {path}"
            )

    def clean_up_env(self, values: Dict[Any, Any]) -> Dict[str, str]:
        """
        returns the environment variable dictionary such that it can be safely used by programs

        :param values: the raw dictionary
        :type values: Dict[Any, Any]
        :raises ConfigError: if a variable name does not comply with POSIX 3.231 Name
        :return: the cleaned up and validated environment variable dictionary
        :rtype: Dict[str, str]
        """
        env = dict()
        pattern = re.compile(r"^[a-zA-Z_]+[a-zA-Z0-9_]*")
        for k, v in values.items():  # type: ignore
            name = str(k)
            if not pattern.match(name):
                raise ConfigError(
                    f"{STARTUP_SCRIPT}: Expected valid environment variable name (see POSIX 3.231 Name)"
                    f" but was '{name}'."
                )
            env[name] = "" if v is None else str(v)
        return env

    def sh(self, environment: Dict[str, str]):
        """
        runs the script with the given environment if the plugin settings demand it

        :param environment: the environment
        :type environment: Dict[str, str]
        :raises IOError: if the script execution fails
        """
        if not self.settings:
            return

        sh = self.settings.sh

        if not sh:
            return

        script = "; ".join(sh)

        extra = {"program": "startup_script/sh"}

        try:
            process = Process(script, environment, shell=True)
            exit_code = process.execute_and_log(lambda _: None, mylogger, extra)

            if exit_code != 0:
                raise IOError(
                    f"{STARTUP_SCRIPT}: Startup script failed with exit code: {exit_code}"
                )
        except BaseException as e:
            raise IOError(f"{STARTUP_SCRIPT}: Failed to execute startup script: {e}")

    def buildenv(self, environment: Dict[str, str]):
        """
        runs the buildenv script, processes and updates the current environment

        :param environment: the environment
        :type environment: Dict[str, str]
        :raises IOError: if the script execution fails
        """

        if not self.settings:
            return

        buildenv = self.settings.buildenv

        if not buildenv:
            return

        script = "; ".join(buildenv)

        extra = {"program": "startup_script/buildenv"}

        mylogger.info("Running buildenv script", extra={"program": ENCAB})

        lines: List[str] = list()
        try:

            def read_lines(process: Popen):
                assert process.stdout
                with process.stdout as stdout:
                    for line in stdout:
                        strline = line.decode(sys.getdefaultencoding()).rstrip(
                            "\r\n\t "
                        )
                        lines.append(strline)

            process = Process(script, environment, shell=True)

            exit_code = process.execute_and_log(
                read_lines, mylogger, extra, log_stdout=False
            )

            if exit_code != 0:
                raise IOError(f"Buildenv script failed with exit code: {exit_code}")

        except BaseException as e:
            raise IOError(f"{STARTUP_SCRIPT}: Failed to execute buildenv script: {e}")

        self.update_env(environment, stream=StringIO("\n".join(lines)))

    def execute(self, environment: Dict[str, str]):
        if self.executed:
            return

        self.executed = True

        self.loadenv(environment)
        self.buildenv(environment)
        self.sh(environment)


extension_impl = HookimplMarker(ENCAB)


class StarupScriptExtension(object):
    def __init__(self) -> None:
        self.script = StartupScript()
        self.enabled = True

    @extension_impl
    def validate_extension(self, name: str, enabled: bool, settings: Dict[str, Any]):
        if name == STARTUP_SCRIPT:
            StartupScriptSettings.load(settings)
            mylogger.info("settings are valid.", extra={"program": ENCAB})

    @extension_impl
    def configure_extension(self, name: str, enabled: bool, settings: Dict[str, Any]):
        if name != STARTUP_SCRIPT:
            return

        if not enabled:
            self.enabled = False
            return

        self.script.update_settings(StartupScriptSettings.load(settings))

    @extension_impl
    def extend_environment(self, program_name: str, environment: Dict[str, str]):
        if not self.enabled:
            return

        self.script.execute(environment)



import os
import pwd

from typing import Dict, Optional, Callable, Any, List, Union

from subprocess import Popen, PIPE
from signal import SIGKILL, SIGTERM

from logging import Logger, INFO, ERROR
from .log_stream import LogStream
from .exit_codes import EX_NOCHILD

from pwd import getpwnam
from grp import getgrnam


def getGroupId(name: str) -> int:
    """
    returns the GID for the group name

    For details see https://docs.python.org/3.10/library/grp.html

    :param name: the group name
    :type name: str
    :return: the GID
    :rtype: int
    :raise: KeyError if the group does not exist
    """
    return getgrnam(name).gr_gid


def getUserId(name: str) -> int:
    """
    returns the UID for the user name

    for details see https://docs.python.org/3/library/pwd.html

    :param name: the user name
    :type name: str
    :return: the UID
    :rtype: int
    :raise: KeyError if the User does not exist
    """
    return getpwnam(name).pw_uid


class Process(object):
    """
    Wrapper for POpen that is backward compatible to Python 3.7.
    In addition, it supports logging of stdout and stderr
    """

    @staticmethod
    def update_current(
        user: Optional[int] = None,
        group: Optional[int] = None,
        umask: Optional[int] = None,
    ):
        """
        sets the current process user, group and umask

        :param user: the UID, defaults to None
        :type user: Optional[int], optional
        :param group: the GID, defaults to None
        :type group: Optional[int], optional
        :param umask: the mask defining the process permssions, defaults to None
        :type umask: Optional[int], optional
        :raises ValueError: The UID cannot be found in the /etc/passwd file
        """

        if group:
            os.setgid(group)

        if user and os.getuid() != user:
            try:
                user_data = pwd.getpwuid(user)
            except KeyError as e:
                raise ValueError(f"No passwd entry for user id: {user}", e)

            os.initgroups(user_data.pw_name, user_data.pw_gid)
            os.setuid(user)

        if umask and umask != -1:
            os.umask(umask)

    def __init__(
        self,
        args: Union[str, List[str]],
        environment: Dict[str, str],
        user: Optional[int] = None,
        group: Optional[int] = None,
        umask: Optional[int] = None,
        shell: bool = False,
        start_new_session: bool = True,
        cwd: Optional[str] = None,
        reap_zombies: bool = False,
    ) -> None:
        """
        Creates the Process

        For details see: https://docs.python.org/3/library/subprocess.html#subprocess.Popen

        :param args: process arguments. The first ist the program to be started
        :type args: Union[str, List[str]]
        :param environment: the process environment
        :type environment: Dict[str, str]
        :param user: the UID for the process , defaults to None
        :type user: Optional[int], optional
        :param group: the GID of the process, defaults to None
        :type group: Optional[int], optional
        :param umask: the process permissions, defaults to None
        :type umask: Optional[int], optional
        :param shell: True: the command is run within a shell, defaults to False
        :type shell: bool, optional
        :param start_new_session: True: the process is executed in its own session, defaults to True
        :type start_new_session: bool, optional
        :param cwd: the work directory, the process is run in, defaults to None
        :type cwd: Optional[str], optional
        :param reap_zombies: if True, encab will reap zombie child processes
        :type reap_zombies: bool, defaults to False
        """
        self._args = args
        self._env = environment
        self._user = user
        self._group = group
        self._umask = umask
        self._shell = shell
        self._start_new_session = start_new_session
        self._process: Optional[Popen] = None
        self._cwd = cwd
        self._reap_zombies = reap_zombies

    def _wait_and_reap_zombies(self, logger: Logger, extra: Dict[str, str]) -> int:
        """
        Waits for the process to end and reap zombies in between

        see: https://github.com/krallin/tini/blob/master/src/tini.c

        :param logger: the logger
        :type logger: Logger
        :param extra: the logger extra
        :type extra: Dict[str, str]
        """

        assert self._process
        child_process_pid = self._process.pid

        while True:
            current_pid, status = os.waitpid(-child_process_pid, os.WUNTRACED)
            if current_pid == -1:
                logger.debug("No child to wait", extra=extra)
                return EX_NOCHILD
            elif current_pid == 0:
                pass
            else:
                if current_pid == child_process_pid:
                    if os.WIFEXITED(status):
                        rc = os.waitstatus_to_exitcode(status)
                        logger.debug(
                            "Main child exited normally (with status %d, exit code %d)",
                            os.WEXITSTATUS(status),
                            rc,
                            extra=extra,
                        )
                        return rc
                    elif os.WIFSIGNALED(status):
                        logger.debug(
                            "Main child exited with signal (with signal '%s')",
                            str(os.WTERMSIG(status)),
                            extra=extra,
                        )
                        return os.waitstatus_to_exitcode(status)
                    else:
                        logger.error(
                            "Main child exited for unknown reason", extra=extra
                        )
                        return os.waitstatus_to_exitcode(status)
                else:
                    logger.warning(
                        "Reaped child with pid: %d", current_pid, extra=extra
                    )

    def execute(
        self,
        exec: Callable[[Popen], Any],
        logger: Logger,
        extra: Dict[str, str],
        stdin: Optional[int] = None,
        stdout: Optional[int] = None,
        stderr: Optional[int] = None,
    ) -> int:
        """
        Executes the process

        For details see:

        - https://docs.python.org/3/library/subprocess.html#subprocess.Popen
        - https://docs.python.org/3.10/library/subprocess.html#replacing-shell-pipeline

        :param exec: function that is called when the process has started
        :type exec: Callable[[Popen], Any]
        :param logger: the logger
        :type logger: Logger
        :param stdin: the stdin file descriptor, defaults to None
        :type stdin: Optional[int], optional
        :param stdout: the stdout file descriptor, defaults to None
        :type stdout: Optional[int], optional
        :param stderr: the stderr file descriptor, defaults to None
        :type stderr: Optional[int], optional
        :raises ValueError: The UID cannot be found in the /etc/passwd file
        :return: the process exit code
        :rtype: int
        """
        uid = self._user
        gid = self._group
        umask = self._umask
        user_data = None

        if uid and os.getuid() != uid:
            try:
                user_data = pwd.getpwuid(uid)
            except KeyError as e:
                raise ValueError(f"No passwd entry for user id: {uid}", e)

        def preexec_fn():
            if gid:
                os.setgid(gid)

            if uid and user_data:
                os.initgroups(user_data.pw_name, user_data.pw_gid)
                os.setuid(uid)

            if umask and umask != -1:
                os.umask(umask)

        self._process = Popen(
            self._args,
            stdin=stdin,
            stdout=stdout,
            stderr=stderr,
            env=self._env,
            preexec_fn=preexec_fn,
            shell=self._shell,
            start_new_session=self._start_new_session,
            cwd=self._cwd,
        )
        exec(self._process)

        if self._reap_zombies:
            return self._wait_and_reap_zombies(logger, extra)
        else:
            self._process.wait()
            return self._process.returncode

    def execute_and_log(
        self,
        exec: Callable[[Popen], Any],
        logger: Logger,
        extra: Any,
        log_stdout: bool = True,
    ) -> int:
        """
        Executes the process and loggs stderr (log level ERROR) and optionally stdout (log level INFO)

        :param exec: function that is called when the process has started
        :type exec: Callable[[Popen], Any]
        :param logger: the logger where stderr and optionally stdout are logget to
        :type logger: Logger
        :param extra: The extra field for log messages
        :type extra: Any
        :param log_stdout: if True, srdout is logged as well, defaults to True
        :type log_stdout: bool, optional
        :return: _description_
        :rtype: int
        """

        streams: List[LogStream] = list()

        def outer_exec(process: Popen) -> None:
            assert process.stdout
            assert process.stderr

            streams.append(LogStream(logger, ERROR, process.stderr, extra).start())

            if log_stdout:
                streams.append(LogStream(logger, INFO, process.stdout, extra).start())

            exec(process)

        try:
            return self.execute(outer_exec, logger, extra, None, PIPE, PIPE)
        finally:
            for stream in streams:
                stream.wait_close()

    def pid(self) -> Optional[int]:
        """
        :return: the operating system process id or None, if the process isn't running.
        :rtype: Optional[int]
        """
        return self._process.pid if self._process and self._process.pid else None

    def if_running(self, f: Callable[[int], Any]) -> Any:
        """
        calls a function if the process is running

        :param f: a function
        :type f: Callable[[int], Any]
        :return: _description_
        :rtype: whatever is ferurned from the function
        """
        return f(self._process.pid) if self._process and self._process.pid else None

    def signal(self, signal: int):
        """
        Sends a singal to the process if it is running

        see: https://docs.python.org/3/library/signal.html#module-signal

        :param signal: the signal to be sent
        :type signal: int
        """
        try:
            self.if_running(lambda pid: os.kill(pid, signal))
        except ProcessLookupError:
            pass

    def kill(self):
        """
        Sends SIGKILL to the process if it is running

        see: https://docs.python.org/3/library/signal.html#module-signal
        """
        self.signal(SIGKILL)

    def terminate(self):
        """
        Sends SIGTERM to the process if it is running

        see: https://docs.python.org/3/library/signal.html#module-signal
        """
        self.signal(SIGTERM)



import sys
import os
import io

from logging import (
    getLogger,
    StreamHandler,
    Formatter,
)

from signal import SIGTERM, SIGINT, signal, getsignal
from typing import Optional, List, Tuple, Dict
from textwrap import shorten
from threading import Event

from .common.process import Process
from .common.exit_codes import (
    EX_OK,
    EX_CONFIG_ERROR,
    EX_IOERROR,
    EX_INSUFFICIENT_PERMISSIONS,
    EX_UNKNOWN_RC,
    EX_INTERRUPTED,
    EX_TERMINATED,
)
from .config import Config, ConfigError
from .program_state import LoggingProgramObserver
from .program import ExecutionContext
from .programs import Programs
from .extensions import extensions, ENCAB

from .ext.log_sanitizer import LogSanitizerExtension
from .ext.validation import ValidationExtension
from .ext.startup_script import StarupScriptExtension
from .ext.log_collector import LogCollectorExtension


def load_config(encab_stream: Optional[io.TextIOBase] = None) -> Tuple[Config, str]:
    """
    Loads the configuration file

    :param encab_stream: _description_, defaults to None
    :type encab_stream: Optional[io.TextIOBase], optional
    :raises FileNotFoundError: _description_
    :return: _description_
    :rtype: Tuple[Config, str]
    """

    encab_file = None

    if encab_stream is not None:
        encab_file = "stream"

    source = "Argument"

    ENCAB_CONFIG = "ENCAB_CONFIG"

    if not encab_file:
        if ENCAB_CONFIG in os.environ:
            encab_file = os.environ[ENCAB_CONFIG]
            source = f"Environment {ENCAB_CONFIG}"

    ENCAB_FILE_CANDIDATES = [
        "./encab.yml",
        "./encab.yaml",
        "/etc/encab.yml",
        "/etc/encab.yaml",
    ]

    if not encab_file:
        for candidate in ENCAB_FILE_CANDIDATES:
            if os.path.exists(candidate):
                encab_file = candidate
                source = "Default location"

    if not encab_file:
        candidates = ", ".join(ENCAB_FILE_CANDIDATES)
        raise FileNotFoundError(f"Encab file not found in {candidates}.")

    if encab_stream:
        config = Config.load(encab_stream)
    else:
        with open(encab_file, "r") as f:
            config = Config.load(f)

    ENCAB_DRY_RUN = "ENCAB_DRY_RUN"

    dry_run = None
    if ENCAB_DRY_RUN in os.environ:
        value = os.environ[ENCAB_DRY_RUN]
        if not value:
            pass
        elif value == "1":
            dry_run = True
        elif value == "0":
            dry_run = False
        else:
            raise ConfigError(
                "Environment variable ENCAB_DRY_RUN"
                " expected to be '1' or '0' if set"
                f" but was '{value}'."
            )

    assert config.encab
    if dry_run is not None:
        config.encab.dry_run = dry_run

    return (config, f"file {encab_file}, source: {source}.")


def set_up_logger(config: Config):
    """
    Sets up the encab logger

    :param config: the encab configuraion
    :type config: Config
    :return: the logger
    :rtype: Logger
    """

    if config.encab:
        root_logger = getLogger()

        handler = StreamHandler()
        formatter = Formatter(config.encab.logformat)

        handler.setFormatter(formatter)

        loglevel = config.encab.loglevel
        assert isinstance(loglevel, int) or isinstance(loglevel, str)

        root_logger.setLevel(loglevel)
        root_logger.addHandler(handler)

    logger = getLogger(ENCAB)
    extensions.update_logger(ENCAB, logger)
    return logger


def set_up_extensions(config: Config, logger, extra: Dict[str, str]):
    """
    sets up and configures the extensions.
    In case of dry run, `encab.extensions.Extensions.validate_extension` instead of
    `encab.extensions.Extensions.configure_extension` is run for each plugin.

    :param config: the encab config
    :type config: Config
    :param logger: the logger to be used
    :type logger: Logger
    :param extra: the logger extension
    :type extra: Dict[str, str]
    """

    if config.extensions:
        assert config.encab and isinstance(config.encab.dry_run, bool)
        dry_run = config.encab.dry_run

        for name, econf in config.extensions.items():
            if econf.module:
                extensions.register_module(econf.module)

            assert isinstance(econf.enabled, bool)

            if dry_run:
                extensions.validate_extension(name, econf.enabled, econf.settings or {})
            else:
                extensions.configure_extension(
                    name, econf.enabled, econf.settings or {}
                )


def encab(
    encab_stream: Optional[io.TextIOBase] = None, args: Optional[List[str]] = None
):
    """
    Encab's main routine

    - loads the configuration file
    - replaces the main program with the one from arguments if specified
    - runs the programs

    :param encab_stream: the configruration as text stream, defaults to None
    :type encab_stream: Optional[io.TextIOBase], optional
    :param args: the main program arguments, defaults to None
    :type args: Optional[List[str]], optional
    """

    extensions.register(
        [
            StarupScriptExtension(),
            LogSanitizerExtension(),
            ValidationExtension(),
            LogCollectorExtension(),
        ]
    )

    logger = None
    try:
        config, location = load_config(encab_stream)

        logger = set_up_logger(config)

        extra = {"program": ENCAB}

        logger.info("encab 1.0.4", extra=extra)
        logger.info("Using configuration %s", location, extra=extra)

        logger.debug(
            "Encab config: %s",
            shorten(str(config), width=127, placeholder="..."),
            extra=extra,
        )

        assert config.encab and isinstance(config.encab.dry_run, bool)
        dry_run = config.encab.dry_run

        if dry_run:
            logger.info("Dry run. No program will be started.", extra=extra)

        set_up_extensions(config, logger, extra)

        if dry_run:
            logger.info("Dry run succeeded. Exiting.", extra=extra)
            return

        config.encab.set_user()
        config.encab.set_group()

        assert config.encab.user is None or isinstance(config.encab.user, int)
        assert config.encab.group is None or isinstance(config.encab.group, int)
        assert config.encab.umask is None or isinstance(config.encab.umask, int)

        Process.update_current(
            config.encab.user, config.encab.group, config.encab.umask
        )

        program_config = config.programs or {}

        if program_config:
            logger.debug("Starting program(s)...", extra=extra)

        observer = LoggingProgramObserver(ENCAB, logger, extra)
        context = ExecutionContext(dict(os.environ), observer)

        if config.encab.environment:
            context = context.extend(config.encab.environment)

        args = args if args is not None else sys.argv[1:]

        programs = Programs(program_config, context, args, config.encab)

        def on_signal(signal, _):
            signames = {SIGINT: "SIGINT", SIGTERM: "SIGTERM"}

            logger.info(
                "Received %s. Interrupting/terminating programs...",
                signames[signal],
                extra=extra,
            )

            if signal == SIGINT:
                programs.interrupt()
                logger.info("Programs interrupted. Exiting.", extra=extra)
                exit(EX_INTERRUPTED)
            else:
                programs.terminate()
                logger.info("Programs terminated. Exiting.", extra=extra)
                exit(EX_TERMINATED)

        sigint_handler = getsignal(SIGINT)
        sigterm_handler = getsignal(SIGTERM)

        signal(SIGINT, on_signal)
        signal(SIGTERM, on_signal)

        programs.run()

        if config.encab.halt_on_exit:
            logger.info("Programs ended. Halting for diagnose.", extra=extra)
            signal(SIGINT, sigint_handler)
            signal(SIGTERM, sigterm_handler)
            Event().wait()
        else:
            logger.debug("Programs ended. Exiting.", extra=extra)

        exit(programs.exit_code if programs.exit_code is not None else EX_UNKNOWN_RC)
    except PermissionError as e:
        print(
            f"Failed to set the encab user: {str(e)}."
            " \nTo set the user, you have to run encab as root."
        )
        exit(EX_INSUFFICIENT_PERMISSIONS)
    except IOError as e:
        print(f"I/O Error: {str(e)}")
        exit((EX_IOERROR))
    except ValueError as e:
        print(f"Error in configuration: {str(e)}")
        exit(EX_CONFIG_ERROR)
    except KeyboardInterrupt:
        print("Encab was interrupted.")
        exit(EX_OK)


if __name__ == "__main__":
    encab()



from copy import deepcopy

from typing import Dict, Optional, Union, List

from subprocess import Popen
from threading import Thread
from signal import SIGINT, SIGTERM

from logging import Logger, getLogger

from .common.process import Process
from .config import ProgramConfig
from .program_state import (
    ProgramObserver,
    ProgramState,
    ProgramStateHandler,
    ProgramCanceledException,
)
from .extensions import extensions


class ExecutionContext(object):
    """
    The context in which a program is executed

    It contains all non-static resources necessary to execute a program,
    such as the logger and the system environment.
    """

    def __init__(self, environment: Dict[str, str], observer: ProgramObserver) -> None:
        self.environment = environment
        self.observer = observer
        extensions.extend_environment(observer.get_name(), environment)

    def extend(self, environment: Dict[str, str]) -> "ExecutionContext":
        env = deepcopy(self.environment)
        if environment:
            env.update(environment)
            extensions.extend_environment(self.observer.get_name(), self.environment)

        return ExecutionContext(env, self.observer)

    def spawn(
        self,
        name: str,
        environment: Dict[str, str],
        logger: Logger,
        extra: Dict[str, str],
    ) -> "ExecutionContext":
        env = deepcopy(self.environment)
        observer = self.observer.spawn(name, logger, extra)

        if environment:
            extensions.extend_environment(name, self.environment)
            env.update(environment)

        extensions.update_logger(name, logger)
        return ExecutionContext(env, observer)


class Program(object):
    """
    Starts a program and captures stdout and stderr as log messages
    """

    def __init__(
        self, name: str, config: ProgramConfig, parent_context: ExecutionContext
    ) -> None:
        """
        :name: (str)
            the program name
        :config: (ProgramConfig)
            the program configuration
        :parent_context (ExecutionContext):
            the parent program execution context
        """
        self.name = name
        self.config = config

        self.logger = getLogger(__name__)

        if config.loglevel:
            self.logger.setLevel(config.loglevel)

        self.extra = {"program": name}
        self.context: ExecutionContext = parent_context.spawn(
            name, self.config.environment or {}, self.logger, self.extra
        )

        if self.config.command:
            self.shell = False
            assert isinstance(self.config.command, list)
            self.command = self.config.command
        else:
            self.shell = True
            assert isinstance(self.config.sh, str)
            self.command = [self.config.sh]

        observer = self.context.observer

        self._observer = observer
        self._state_handler = ProgramStateHandler(observer)
        self._process: Optional[Process] = None
        self.exit_code: Optional[int] = None

    def _run(self) -> None:
        logger = self.logger
        extra = self.extra

        command = self.command
        env = self.context.environment
        state = self._state_handler
        observer = self._observer

        startup_delay = self.config.startup_delay

        umask = self.config.umask
        user = self.config.user
        group = self.config.group
        cwd = self.config.directory
        reap_zombies = self.config.reap_zombies
        restart_delay = self.config.restart_delay

        assert user is None or isinstance(user, int)
        assert group is None or isinstance(group, int)
        assert isinstance(reap_zombies, bool)

        try:
            assert isinstance(startup_delay, float) or isinstance(startup_delay, int)
            assert isinstance(umask, int)

            state.wait(float(startup_delay))

            observer.on_execution(command, env, self.config)
            state.set(ProgramState.STARTING)

            shell = self.shell
            args: Union[str, List[str]] = command

            if shell:
                args = command[0]

            def on_run(popen: Popen):
                state.set(ProgramState.RUNNING)
                observer.on_run(popen.pid)

            while True:
                self._process = Process(
                    args,
                    env,
                    user=user,
                    group=group,
                    umask=umask,
                    shell=shell,
                    start_new_session=True,
                    cwd=cwd,
                    reap_zombies=reap_zombies,
                )

                self.exit_code = self._process.execute_and_log(on_run, logger, extra)
                state.handle_exit(self.exit_code, self.command)

                if restart_delay is not None:
                    state.set(ProgramState.STARTING)
                    state.wait(float(restart_delay))
                else:
                    break

        except ProgramCanceledException:
            observer.on_cancel()
            state.set(ProgramState.CANCELED)
        except BaseException as e:
            observer.on_crash(self.command, e)
            state.set(ProgramState.CRASHED)

    def get_state(self) -> int:
        return self._state_handler.get()

    def start(self, timeout: Optional[float] = 1) -> int:
        thread = Thread(target=lambda: self._run(), name=self.name)
        thread.daemon = True
        thread.start()

        self._observer.on_start()

        return self._state_handler.wait_for_startup(timeout)

    def join(
        self,
        timeout: Optional[float] = None,
    ) -> ProgramState:
        return self._state_handler.join(timeout)

    def interrupt(self) -> None:
        return self._state_handler.signal(self._process, SIGINT)

    def terminate(self) -> None:
        return self._state_handler.signal(self._process, SIGTERM)

    def join_wait(self, timeout: Optional[float] = None) -> ProgramState:
        return self._state_handler.join_wait(timeout)



"""
Linux /usr/include/sysexits.h
"""

EX_OK = 0  # successful termination
EX_DATAERR = 65  # data format error
EX_NOINPUT = 66  # cannot open input
EX_UNAVAILABLE = 69  # service unavailable
EX_SOFTWARE = 70  # internal software error
EX_OSERR = 71  # system error (e.g., can't fork) */
EX_IOERR = 74  # input/output error */
EX_NOPERM = 77  # permission denied
EX_CONFIG = 78  # configuration error

EX_NOCHILD = EX_OSERR  # Child process could not be launched
EX_UNKNOWN_RC = EX_OSERR  # Return code of main program could not be determined
EX_IOERROR = EX_IOERR  # I/O Error, e.g. reading config
EX_CONFIG_ERROR = EX_CONFIG  # Configuration error
EX_INSUFFICIENT_PERMISSIONS = EX_NOPERM  # Insufficient permissions to set UID or GID
EX_INTERRUPTED = EX_OK  # Encab was interrupted
EX_TERMINATED = EX_OK  # Encab was terminated
